local fps = 1/40;
if not LibStub then return end
local libxml = LibStub("LibXML-1.0");

local function Engine_OnKeyDown(self, caller, key)
	if not self.protagonist then return end
	if key == "P" then
		if self.running then
			self.running = false;
			self.backdrop:SetTexture(0.2,0.2,0.3,0.4);
			self.caption:SetText("Paused!");
		else
			self.running = true;
			self.backdrop:SetTexture(nil);
			self.caption:SetText("");
		end
	end
	if not self.running then return end
	if 		key == "LEFT" then
		self.protagonist.direction = "left"
		self.keys.leftDown = true;
		self.keys.rightDown = false;
	elseif	key == "RIGHT" then
		self.protagonist.direction = "right"
		self.keys.leftDown = false;
		self.keys.rightDown = true;
	elseif key == "UP" then
		self.keys.jumpFired = true;
	elseif key == "SPACE" then
		self.keys.spaceFired = true;
	end
end

local function Engine_OnKeyUp(self, caller, key)
	if not self.protagonist then return end
	if 		key == "LEFT" then
		self.keys.leftDown = false;
	elseif	key == "RIGHT" then
		self.keys.rightDown = false;
	elseif key == "SPACE" then
		self.keys.spaceFired = false;
	elseif key == "UP" then
		self.keys.jumpFired = false;
	end
end

local function Engine_SetSpeed(self, speed)
	self.gameSpeed = tonumber(speed) or self.gameSpeed;
end

local function Engine_Update(self, ...)
	local speed = 0;
	if not self.running then return end
	if self.keys.leftDown == true then speed = -(self.protagonist.speed); end
	if self.keys.rightDown == true then speed = (self.protagonist.speed); end
	self.protagonist.forceH = speed;
	local now = GetTime();
	if (now - self.lastUpdate) >= (fps) then
		if self.keys.jumpFired and self.protagonist:OnGround() then
			self.protagonist:Jump()
			self.keys.jumpFired = false;
		end
		self.world:Step(1/self.gameSpeed);
		for n = 1, #self.world.creatures do
			if self.world.creatures[n].script then
				self.world.creatures[n]:script();
			end
		end
		self.lastUpdate = now;
	end
end

local function Engine_CallBack(t, mob1, side1, mob2, side2)
	if mob1.static == true then
		return "BLOCK", mob2
	elseif mob2.static == true then
		return "BLOCK", mob1
	end
	return "IGNORE"
end

local function Engine_LoadWorld(self, lid)
	if not self.world then
		self.world = Gnomario_World_Create(self.frame,480,320, self.CallBack, self.xml);
	end
	self.world:Clear();
	for n = 1, #self.xml.levels do
		local level = self.xml.levels[n];
		if level.args and level.args.id == lid then
			for i = 1, #level do
				if level[i].class then
					local o = level[i];
					for key, obj in pairs(self.xml.environment or {}) do
						if type(obj) == "table" and obj.args and obj.args.id == o.class then
							local x = self.world:Spawn_Object(obj.args.id);
							x.x = tonumber(o.args.x) or 0;
							x.y = tonumber(o.args.y) or 0;
							break;
						end
					end
					for key, obj in pairs(self.xml.creatures or {}) do
						if type(obj) == "table" and obj.args and obj.args.id == o.class then
							local x = self.world:Spawn_Creature(obj.args.id);
							x.x = tonumber(o.args.x) or 0;
							x.y = tonumber(o.args.y) or 0;
							break;
						end
					end
				end
			end
			break;
		end
	end
end

local function Engine_Start(self)
	self.running = true;
	self.lastUpdate = GetTime();
end

local function Engine_SetProtagonist(self, mob)
	self.protagonist = mob;
end

function Gnomario_Engine(parent, xml)
	local engine = {};
	engine.sprites = {};
	engine.keys = {};
	local xml_data, err = libxml:Import(xml);
	if err then error(err); return end
	engine.xml = {};
	engine.xml.scripts = engine.xml.scripts or {};
	for n=1, #xml_data[1] do
		if xml_data[1][n].class == "environment" then
			engine.xml.environment = xml_data[1][n];
		elseif xml_data[1][n].class == "creatures" then
			engine.xml.creatures = xml_data[1][n];
		elseif xml_data[1][n].class == "levels" then
			engine.xml.levels = xml_data[1][n];
		elseif xml_data[1][n].class == "scripts" then
			local scripts = xml_data[1][n];
			for m = 1, #scripts do
				local script = scripts[m];
				if type(script) == "table" and script.args.id then
					local txt = "";
					for x = 1, #script do
						if type(script[x]) == "string" then
							txt = txt .. script[x];
						end
					end
					engine.xml.scripts[script.args.id],err = loadstring("do local self = ...;" .. txt .. "end");
					if err then
						error("Error while parsing game script [" .. script.args.id .."] : " .. err);
					end
				end
			end
		end
	end
	engine.frame = CreateFrame("Frame");
	engine.frame:SetParent(parent);
	engine.frame:EnableKeyboard(true);
	engine.frame:SetAllPoints(parent);
	engine.frame:SetScript("OnKeyDown", function(...) Engine_OnKeyDown(engine,...) end);
	engine.frame:SetScript("OnKeyUp", function(...) Engine_OnKeyUp(engine,...) end);
	engine.frame:SetScript("OnUpdate", function(...) Engine_Update(engine, ...) end);
	engine.backdrop = engine.frame:CreateTexture("OVERLAY");
	engine.backdrop:SetAllPoints(engine.frame);
	-- caption
	local caption = engine.frame:CreateFontString(nil, "DIALOG");
	caption:SetParent(engine.frame);
	caption:SetFontObject(ChatFontNormal);
	caption:SetText("");
	caption:SetPoint("TOPLEFT", engine.frame, "TOPLEFT", 5, -5);
	caption:SetPoint("BOTTOMRIGHT", engine.frame, "TOPRIGHT", 5, -25);
	caption:SetTextColor(0.9,0.5,0.2,1);
	engine.caption = caption;

	engine.gameSpeed = 25;
	engine.running = false;
	engine.SetSpeed = Engine_SetSpeed;
	engine.LoadWorld = Engine_LoadWorld;
	engine.Start = Engine_Start;
	engine.SetProtagonist = Engine_SetProtagonist;
	engine.CallBack = Engine_CallBack;
	return engine;
end
