local g = 9.81;
local ppm = 57; -- pixels per meter.

local function Gnomario_World_AtEdge(self)
	return self.walled
end

local function Gnomario_World_OnGround(self)
	return self.grounded;
end
local function Gnomario_World_AtWall(self)
	return (self.grounded and self.walled);
end
local function Gnomario_World_Jump(self)
	if self.grounded then
		self.forceV = 5.35;
		PlaySoundFile([[Interface\AddOns\Gnomario\Sounds\jump.ogg]]);
	end
end


local function Apply_Physics(self)
	self.AtEdge = Gnomario_World_AtEdge;
	self.OnGround = Gnomario_World_OnGround;
	self.AtWall = Gnomario_World_AtWall;
	self.Jump = Gnomario_World_Jump;
end

local function Gnomario_Physics_Add(self, object)
	table.insert(self.objects, object);
end

local function Gnomario_Physics_Step(self, duration)
	for n = 1, #self.objects do
		local collided = false;
		local obj = self.objects[n];
		local x,y = obj.x, obj.y;
		if not obj.static then
			obj.offsetH = obj.offsetH or 0;
			obj.forceV = obj.forceV or 0;
			if obj.forceV ~= 0 then
				obj.grounded = false;
			end
			obj.forceV = (obj.forceV or 0) - ( (obj.g or g) * duration );
			if (obj.forceH < 0 and obj.limitLeft) or (obj.forceH > 0 and obj.limitRight) then obj.forceH = 0; end
			if (obj.offsetH < 0 and obj.limitLeft) or (obj.offsetH > 0 and obj.limitRight) then obj.offsetH = 0; end
			if (obj.forceV < 0 and obj.limitDown) or (obj.forceV > 0 and obj.limitUp) then obj.forceV = 0; end
			y = y + (obj.forceV  * duration * ppm);

			if y <= 0 or (obj.landed and y <= (obj.landed.y + obj.landed.height )) then
				y = math.max(0, y);
				if obj.landed then
					y = obj.landed.y + obj.landed.height;
				end
				obj.grounded = true;
				obj.forceV = 0;
			end
			if ( obj.grounded or true ) then
				x = x + ( (obj.forceH or 0) * duration * ppm )
				obj.offsetH = obj.forceH;
			else
				x = x + ( (obj.offsetH or 0) * duration * ppm )
			end
			-- bla bla bla collision test goes here
			obj.limitLeft, obj.limitRight, obj.limitUp, obj.limitDown = false, false, false, false;
			local x1, y1 = x + (obj.width/2), y + (obj.height/2);
			for m = 1, #self.objects do
				local cobj = self.objects[m];
				if cobj ~= obj then
					local x2, y2 = cobj.x + (cobj.width/2), cobj.y + (cobj.height/2);
					local dx, dy = math.abs(x2-x1), math.abs(y2-y1);
					if ( dx <= (cobj.width + obj.width)/2 ) and ( dy <= (cobj.height + obj.height)/2 ) then
						collided = true;
						if ( dx <= dy ) then
							if obj.y >= cobj.y then
								t = {'landed', obj, 'bottom', cobj, 'top'}
							else
								t = {'landed', cobj, 'bottom', obj, 'top'}
							end
						else
							if obj.x >= cobj.x then
								t = {'walled', obj, 'left', cobj, 'right'}
							else
								t = {'walled', cobj, 'left', obj, 'right'}
							end
						end
--						if obj.collider ~= cobj then
							--print(GetTime(), ": collision!", unpack(t));
							if type(self.callback) == "function" then
								local ret, who = self.callback(unpack(t));
								--print("Callback: ", ret,who);
								if ret == "BLOCK" then
									if t[1] == 'walled' then
										if who == obj then
											who.walled = cobj;
											if x <= cobj.x then who.limitRight = true; else who.limitLeft = true; end
										else
											who.walled = obj;
											if x <= obj.x then who.limitRight = true; else who.limitLeft = true; end
										end
									else
										who.forceV = 0;
										if who == obj then
											--who.landed = cobj;
											if who.y < cobj.y then who.limitUp = true; else who.limitDown = true; y = cobj.y + cobj.height;who.landed = cobj; who.grounded = true;end
										else
											--who.landed = obj;
											if who.y < obj.y then who.limitUp = true; else who.limitDown = true; y = obj.y + obj.height; who.landed = cobj; who.grounded = true;end
										end
									end
								elseif ret == "IGNORE" then
									if t[1] == 'walled' then
										obj.walled = nil;
									end
								end
							end
--						end
						obj.collider = cobj;
						cobj.collider = obj;
					end
				end
			end
			if not collided then
				obj.collider = nil;
				obj.walled = nil;
				obj.landed = nil;
				obj.limitLeft, obj.limitRight, obj.limitUp, obj.limitDown = false, false, false, false;
			end
			if not ((obj.limitLeft and x <= obj.px) or (obj.limitRight and x >= obj.px)) then
				obj.x = x;
				obj.px = x;
			end
			if not (obj.limitUp and y > obj.py) or (obj.limitDown and y < obj.py) then
				obj.y = y;
				obj.py = y;
			end
			local t = {};
			if obj.limitLeft then tinsert(t, 'left') end
			if obj.limitRight then tinsert(t, 'right') end
			if obj.limitUp then tinsert(t, 'up') end
			if obj.limitDown then tinsert(t, 'down') end
			if #t == 0 then t = {'none'} end
			if not obj.hostile then
				Gnomario.engine.caption:SetText(("x: %d, y: %d, locks: %s - ground: %s - h-speed: %.2fm/s v-speed: %.2fm/s"):format(x,y,table.concat(t,', '),tostringall(obj.grounded), obj.forceH, obj.forceV));
			end
		end
		--obj.x,obj.y = x,y;
	end
end

local function Gnomario_Physics_Put(self, obj, x, y)
	obj.x = x;
	obj.y = y;
	obj.px = x;
	obj.py = y;
	obj.limitLeft, obj.limitRight, obj.limitUp, obj.limitDown = false, false, false, false;
	obj.forceH, obj.forceV = 0,0;
	Apply_Physics(obj);
	table.insert(self.objects, obj);
end

function Gnomario_Physics_Chamber(width,height,callback)
	local chamber = {};
	chamber.width = width;
	chamber.height = height;
	chamber.objects = {};
	chamber.callback = callback;
	chamber.Step = Gnomario_Physics_Step;
	chamber.Add = Gnomario_Physics_Add;
	chamber.Put = Gnomario_Physics_Put;
	return chamber;
end
